import json
import uuid

from gpt_image.geometry import Geometry
from gpt_image.partition import (Partition, PartitionAttribute,
                                 PartitionEntryArray)

PART_NAME = "test-part"
PART_NAME_2 = "partition-2"
PART_UUID = "26be6d04-85fe-4fae-ba9c-1f47cf16f8d8"


def test_partition_init_guid():
    part = Partition(PART_NAME, 2 * 1024, Partition.LINUX_FILE_SYSTEM)
    assert isinstance(uuid.UUID(part.partition_guid), uuid.UUID)
    del part
    part = Partition(PART_NAME, 2 * 1024, Partition.LINUX_FILE_SYSTEM, PART_UUID)
    assert part.partition_guid == PART_UUID


def test_partition_repr():
    part = Partition(PART_NAME, 2 * 1024, Partition.LINUX_FILE_SYSTEM)
    part.attribute_flags = PartitionAttribute.READ_ONLY
    part_s = str(part)
    assert PART_NAME in part_s
    # attributes with leading underscore should not be in __repr__
    assert "_attribute" not in part_s
    part_d = json.loads(part_s)
    assert part_d.get("partition_name") == PART_NAME
    assert part_d.get("attribute_flags") == [PartitionAttribute.READ_ONLY.value]


def test_partition_attribute():
    part = Partition(PART_NAME, 2 * 1024, Partition.LINUX_FILE_SYSTEM)
    # flags are set to NONE by default
    assert part.attribute_flags == []

    # setting the READ_ONLY flag sets bit 60 which will have an
    # integer value
    part.attribute_flags = PartitionAttribute.READ_ONLY
    assert part.attribute_flags == [PartitionAttribute.READ_ONLY.value]
    part.attribute_flags = PartitionAttribute.HIDDEN
    # this should return 2 values, HIDDEN and READ_ONLY
    assert part.attribute_flags == [
        PartitionAttribute.HIDDEN.value,
        PartitionAttribute.READ_ONLY.value,
    ]

    # setting NONE should clear all flags
    part.attribute_flags = PartitionAttribute.NONE
    assert part.attribute_flags == []


def test_partition_marshal():
    part = Partition(
        PART_NAME, 2 * 1024, Partition.LINUX_FILE_SYSTEM, partition_guid=PART_UUID
    )
    part.attribute_flags = PartitionAttribute.READ_ONLY
    part_bytes = part.marshal()
    assert part_bytes[0:16] == uuid.UUID(Partition.LINUX_FILE_SYSTEM).bytes_le
    assert part_bytes[16:32] == uuid.UUID(PART_UUID).bytes_le
    assert part_bytes[32:40] == b"\x00" * 8
    assert part_bytes[40:48] == b"\x00" * 8
    assert part_bytes[48:56] == (2**PartitionAttribute.READ_ONLY.value).to_bytes(
        8, byteorder="little"
    )
    # pad this to 72 bytes
    assert part_bytes[56:128] == bytes(PART_NAME, encoding="utf_16_le").ljust(
        72, b"\x00"
    )


def test_partition_read():
    geo = Geometry(8 * 1024 * 1024)
    part_array = PartitionEntryArray(geo)
    part = Partition(
        PART_NAME, 2 * 1024, Partition.LINUX_FILE_SYSTEM, partition_guid=PART_UUID
    )
    part_array.add(part)
    part_bytes = part.marshal()
    new_part = Partition.read(part_bytes, geo.sector_size)
    assert new_part.partition_name == PART_NAME
    assert new_part.size == 2 * 1024
    assert new_part.type_guid.upper() == Partition.LINUX_FILE_SYSTEM
    assert new_part.partition_guid == PART_UUID
    # @TODO: read attributes from existing partitions


def test_partition_entry_add():
    geo = Geometry(8 * 1024 * 1024)
    part_array = PartitionEntryArray(geo)
    part1 = Partition(PART_NAME, 2 * 1024, Partition.LINUX_FILE_SYSTEM)
    part_array.add(part1)
    # check partition LBAs. The default alignment is 8 sectors so the
    # first LBA will always be a factor of 8
    assert part1.first_lba == 40
    assert part1.last_lba == 43
    part2 = Partition(PART_NAME_2, 3 * 1024, Partition.LINUX_FILE_SYSTEM)
    part_array.add(part2)
    assert part2.first_lba == 48
    assert part2.last_lba == 53

    assert len(part_array.entries) == 2


def test_get_first_lba():
    pass


def test_get_last_lba():
    pass


def test_partition_entry_marshall():
    geo = Geometry(8 * 1024 * 1024)
    part_array = PartitionEntryArray(geo)
    part1 = Partition(PART_NAME, 2 * 1024, Partition.LINUX_FILE_SYSTEM)
    part2 = Partition(PART_NAME_2, 3 * 1024, Partition.LINUX_FILE_SYSTEM)
    part_array.add(part1)
    part_array.add(part2)

    part_array_bytes = part_array.marshal()
    assert (
        len(part_array_bytes)
        == PartitionEntryArray.EntryCount * PartitionEntryArray.EntryLength
    )
    part1_bytes = part_array_bytes[:128]
    part2_bytes = part_array_bytes[128:256]
    part1_unmarshal = Partition.read(part1_bytes, geo.sector_size)
    part2_unmarshal = Partition.read(part2_bytes, geo.sector_size)
    assert part1_unmarshal.partition_name == PART_NAME
    assert part2_unmarshal.partition_name == PART_NAME_2



import json
import uuid

import pytest

from gpt_image.geometry import Geometry
from gpt_image.table import Header, ProtectiveMBR, Table

DISK_SIZE = 2 * 1024 * 1024  # 2 MB disk
DISK_GUID = "51581dee-faaa-404d-b0fe-85dc65157702"


@pytest.fixture
def new_geometry():
    return Geometry(DISK_SIZE)


def test_protective_mbr_marshall(new_geometry: Geometry):
    pmbr = ProtectiveMBR(new_geometry)
    pmbr_bytes = pmbr.marshal()
    assert pmbr_bytes[:446] == b"\x00" * 446
    assert pmbr_bytes[446:447] == b"\x00"
    assert pmbr_bytes[447:450] == b"\x00" * 3
    assert pmbr_bytes[450:451] == b"\xEE"
    assert pmbr_bytes[451:454] == b"\x00" * 3
    assert pmbr_bytes[454:458] == b"\x01\x00\x00\x00"
    assert pmbr_bytes[458:462] == b"\xff\x0f\x00\x00"
    assert pmbr_bytes[462:510] == b"\x00" * 48
    assert pmbr_bytes[510:512] == b"\x55\xAA"


def test_protective_mbr_read(new_geometry: Geometry):
    pmbr = ProtectiveMBR(new_geometry)
    pmbr_bytes = pmbr.marshal()

    pmbr_o = ProtectiveMBR.read(pmbr_bytes, new_geometry)
    assert pmbr_o.boot_indicator == 0
    assert pmbr_o.start_chs == b"\x00"
    assert pmbr_o.partition_type == b"\xEE"
    assert pmbr_o.end_chs == b"\x00"
    assert pmbr_o.start_sector == 1
    assert pmbr_o.partition_size == 4095
    assert pmbr_o.signature == b"\x55\xAA"


def test_header_marshall(new_geometry: Geometry):
    header = Header(
        new_geometry,
        guid=DISK_GUID,
    )
    header_b = header.marshal()
    assert len(header_b) == new_geometry.sector_size
    assert header_b[:8] == b"EFI PART"
    assert header_b[8:12] == b"\x00\x00\x01\x00"
    assert header_b[12:16] == b"\\\x00\x00\x00"
    assert header_b[16:20] == b"\x00" * 4
    assert header_b[20:24] == b"\x00" * 4
    assert header_b[24:32] == b"\x01" + b"\x00" * 7
    assert header_b[32:40] == b"\xff\x0f" + b"\x00" * 6
    assert header_b[40:48] == b'"\x00' + b"\x00" * 6
    assert header_b[48:56] == b"\xde\x0f" + b"\x00" * 6
    assert header_b[56:72] == uuid.UUID(DISK_GUID).bytes_le
    assert header_b[72:80] == b"\x02" + b"\x00" * 7
    assert header_b[80:84] == b"\x80" + b"\x00" * 3
    assert header_b[84:88] == b"\x80" + b"\x00" * 3
    assert header_b[88:92] == b"\x00" * 4


def test_header_read(new_geometry: Geometry):
    header = Header(
        new_geometry,
        guid=DISK_GUID,
    )
    header_b = header.marshal()
    assert len(header_b) == new_geometry.sector_size

    header_o = Header.read(header_b[: Header._HEADER_SIZE], new_geometry)
    assert header_o.signature == Header._SIGNATURE
    assert header_o.revision == Header._REVISION
    assert header_o.header_size == Header._HEADER_SIZE


def test_header_repr(new_geometry: Geometry):
    header = Header(new_geometry, guid=DISK_GUID)
    header_s = header.__repr__()
    assert "EFI PART" in header_s
    header_d = json.loads(header_s)
    assert header_d.get("header_size") == Header._HEADER_SIZE
    assert len(header_d) == 15


def test_table_init(new_geometry):
    geo = new_geometry
    table = Table(geo)
    assert table.primary_header.backup is False
    assert table.secondary_header.backup is True


def test_checksum_partitions(new_geometry):
    table = Table(new_geometry)
    # partition checksum will be blank before when initialized
    assert table.primary_header.partition_entry_array_crc32 == 0
    table.checksum_partitions(table.primary_header)
    # test that the checksum is no longer zeroed
    assert (
        table.primary_header.partition_entry_array_crc32 == 2874462854
    )  # predictable binascii.crc32


def test_checksum_header(new_geometry):
    table = Table(new_geometry)
    assert table.primary_header.header_crc32 == 0
    # create a predictable GUID so that the test is predictable
    table.primary_header.disk_guid = DISK_GUID
    table.checksum_header(table.primary_header)
    assert table.primary_header.header_crc32 == 82915359


def test_update(new_geometry):
    table = Table(new_geometry)
    table.update()
    assert table.primary_header.header_crc32 != b"\x00" * 4
    assert table.secondary_header.header_crc32 != b"\x00" * 4
    assert table.primary_header.partition_entry_array_crc32 != b"\x00" * 4
    assert table.secondary_header.partition_entry_array_crc32 != b"\x00" * 4



import json
import shutil
import subprocess
import sys
import uuid

import pytest
from gpt_image import disk
from gpt_image.partition import Partition, PartitionAttribute

STATIC_UUID = "4133e7fe-0be9-4097-b617-e3373fa0535e"
DISK_SIZE = 2 * 1024 * 1024  # (2MB)
PART1_NAME = "partition1"
PART1_SIZE = 2 * 1024  # 2KB
PART2_NAME = "partition-number2"
PART2_SIZE = 4 * 1024  # 4KB


@pytest.fixture
def create_image(tmp_path):
    image = tmp_path / "test-disk.raw"

    # this should produce a disk that can be opened and validated with GPT tools
    new_disk = disk.Disk(image)
    new_disk.create(DISK_SIZE)
    part1 = Partition(PART1_NAME, PART1_SIZE, Partition.LINUX_FILE_SYSTEM)
    part1.attribute_flags = PartitionAttribute.HIDDEN
    new_disk.table.partitions.add(part1)

    part2 = Partition(PART2_NAME, PART2_SIZE, Partition.LINUX_FILE_SYSTEM, STATIC_UUID)
    new_disk.table.partitions.add(part2)

    new_disk.write()
    return image


@pytest.mark.skipif(sys.platform != "linux", reason="requires linux to run")
@pytest.mark.skipif(shutil.which("sfdisk") is None, reason="requires sfdisk utility")
def test_sfdisk(create_image):
    sfdisk = shutil.which("sfdisk")
    result = subprocess.run(
        [sfdisk, "--json", create_image], capture_output=True, text=True
    )
    assert result.returncode == 0
    part_info = json.loads(result.stdout)
    assert part_info.get("partitiontable").get("label") == "gpt"
    # a bit of math is required to get the total disk size from the sfdisk output
    # the 34 sectors of GPT header must be added to the lastlba to get the total size
    size = (part_info.get("partitiontable").get("lastlba") + 34) * 512
    assert size == DISK_SIZE

    # test partitions
    partitions = part_info.get("partitiontable").get("partitions")
    assert partitions[0].get("name") == PART1_NAME
    assert partitions[0].get("size") == (PART1_SIZE / 512)
    assert (
        partitions[0].get("start") == 40
    )  # 34 is the first usable but alignment is 8, so 40 is the start
    assert partitions[0].get("attrs") == "GUID:62"
    assert partitions[1].get("name") == PART2_NAME
    assert partitions[1].get("size") == (PART2_SIZE / 512)
    assert partitions[1].get("start") == 48
    assert partitions[1].get("uuid") == STATIC_UUID.upper()



"""
Header information reference: https://en.wikipedia.org/wiki/GUID_Partition_Table

"""
import binascii
import json
import struct
import uuid

from gpt_image.geometry import Geometry
from gpt_image.partition import PartitionEntryArray


class HeaderReadError(Exception):
    """Exception from reading existing headers"""


class ProtectiveMBR:
    """Protective MBR Table Entry

    Provides the bare minimum entries needed to represent a protective MBR.
    https://thestarman.pcministry.com/asm/mbr/PartTables.htm#pte

    """

    _MBR_FORMAT = struct.Struct("<446sB3sc3sII48s2s")

    def __init__(
        self,
        geometry: Geometry,
        partition_type: bytes = b"\xEE",
        signature: bytes = b"\x55\xAA",
    ):
        self._geometry = geometry
        self._start_padding = b"\x00"  # padding before boot indicator
        self.boot_indicator = 0  # not bootable
        self.start_chs = b"\x00"  # ignore the start CHS
        self.partition_type = partition_type  # GPT partition type
        self.end_chs = b"\x00"  # ignore the end CHS
        self.start_sector = self._geometry.my_lba
        # size, minus the protective MBR sector. This only works if the
        # disk is under 2 TB
        self.partition_size = int(self._geometry.total_sectors - 1)
        self._end_padding = b"\x00"  # padding before signature
        self.signature = signature

    def marshal(self) -> bytes:
        """Convert Protective MBR to its byte structure"""
        pmbr_bytes = self._MBR_FORMAT.pack(
            self._start_padding,
            self.boot_indicator,
            self.start_chs,
            self.partition_type,
            self.end_chs,
            self.start_sector,
            self.partition_size,
            self._end_padding,
            self.signature,
        )
        return pmbr_bytes

    @staticmethod
    def read(pmbr_bytes: bytes, geometry: Geometry) -> "ProtectiveMBR":
        (
            _,
            boot_indicator,
            start_chs,
            partition_type,
            end_chs,
            start_sector,
            partition_size,
            _,
            signature,
        ) = ProtectiveMBR._MBR_FORMAT.unpack(pmbr_bytes)

        return ProtectiveMBR(geometry, partition_type, signature)


class Header:
    """GPT Partition Table Header Object

    Each table has two GPT headers, primary and secondary (backup). The primary is
    written to LBA 1 and secondary is written to LBA -1. The GPT Headers contain
    various data including the locations of one another. Therefore two headers
    are created for each table.  The only common component is the disk GUID.

    """

    _HEADER_FORMAT = struct.Struct("<8s4sIIIQQQQ16sQIII")
    _SIGNATURE = b"EFI PART"
    _REVISION = b"\x00\x00\x01\x00"
    _HEADER_SIZE = 92

    def __init__(
        self,
        geometry: Geometry,
        header_crc32: int = 0,
        partition_entry_array_crc32: int = 0,
        guid: str = "",
        is_backup: bool = False,
    ):
        self._geometry = geometry
        self.backup = is_backup
        # header fields
        self.signature = self._SIGNATURE
        self.revision = self._REVISION
        self.header_size = self._HEADER_SIZE
        self.header_crc32 = header_crc32
        self.reserved = 0
        self.my_lba = self._geometry.my_lba
        self.alternate_lba = self._geometry.alternate_lba
        self.first_usable_lba = self._geometry.first_usable_lba
        self.last_usable_lba = self._geometry.last_usable_lba
        self.disk_guid = guid
        # create a GUID if it's a blank string
        if not guid:
            self.disk_guid = str(uuid.uuid4())
        self.partition_entry_lba = self._geometry.partition_entry_lba
        self.number_of_partition_entries = 128
        self.size_of_partition_entries = 128
        self.partition_entry_array_crc32 = partition_entry_array_crc32

        if self.backup:
            self.my_lba, self.alternate_lba = (
                self.alternate_lba,
                self.my_lba,
            )
            self.partition_entry_lba = self._geometry.alternate_array_lba

    def __repr__(self) -> str:
        filtered_values = {k: v for k, v in vars(self).items() if not k.startswith("_")}
        # convert dictionary values that are bytes to strings
        converted_values = {
            k: v.decode() for k, v in filtered_values.items() if isinstance(v, bytes)
        }
        filtered_values.update(converted_values)
        return json.dumps(filtered_values, indent=2, ensure_ascii=False)

    def marshal(self) -> bytes:
        header_bytes = self._HEADER_FORMAT.pack(
            self.signature,
            self.revision,
            self.header_size,
            self.header_crc32,
            self.reserved,
            self.my_lba,
            self.alternate_lba,
            self.first_usable_lba,
            self.last_usable_lba,
            uuid.UUID(self.disk_guid).bytes_le,
            self.partition_entry_lba,
            self.number_of_partition_entries,
            self.size_of_partition_entries,
            self.partition_entry_array_crc32,
        )
        assert len(header_bytes) == self._HEADER_SIZE
        padding = b"\x00" * (self._geometry.sector_size - len(header_bytes))
        return header_bytes + padding

    @staticmethod
    def read(header_bytes: bytes, geometry: Geometry) -> "Header":
        (
            signature,
            revision,
            header_size,
            header_crc32,
            reserved,
            my_lba,
            alternate_lba,
            first_usable_lba,
            last_usable_lba,
            disk_guid,
            partition_entry_lba,
            number_of_partitions,
            size_of_partitions,
            partition_entry_crc32,
        ) = Header._HEADER_FORMAT.unpack(header_bytes)

        if signature != Header._SIGNATURE:
            raise HeaderReadError("Invalid GPT header signature")

        if revision != Header._REVISION:
            raise HeaderReadError("Invalid GPT revision")

        if header_size != Header._HEADER_SIZE:
            raise HeaderReadError("Invalid header size")

        geometry.my_lba = my_lba
        geometry.alternate_lba = alternate_lba
        geometry.first_usable_lba = first_usable_lba
        geometry.last_usable_lba = last_usable_lba
        geometry.partition_entry_lba = partition_entry_lba

        return Header(
            geometry,
            header_crc32,
            partition_entry_crc32,
            disk_guid,
        )


class Table:
    """GPT Partition Table Object

    The entire GPT table structure including the protective MBR,
    GPT headers and partition tables
    """

    def __init__(self, geometry: Geometry):
        self._geometry = geometry
        self.protective_mbr: ProtectiveMBR = ProtectiveMBR(self._geometry)
        self.primary_header: Header = Header(self._geometry)
        self.secondary_header: Header = Header(self._geometry, is_backup=True)
        self.partitions: PartitionEntryArray = PartitionEntryArray(self._geometry)

    def update(self) -> None:
        # calculate partition checksum and write to header
        self.checksum_partitions(self.primary_header)
        self.checksum_partitions(self.secondary_header)

        # calculate header checksum and write to header
        self.checksum_header(self.primary_header)
        self.checksum_header(self.secondary_header)

    def checksum_partitions(self, header: Header) -> None:
        """Checksum the partition entries

        Args:
            header: initialized GPT header object
        """
        part_entry_bytes = self.partitions.marshal()
        header.partition_entry_array_crc32 = binascii.crc32(part_entry_bytes)

    def checksum_header(self, header: Header) -> None:
        """Checksum the table header

        This CRC includes the partition checksum, and must be calculated
        after that has been written.

        Args:
            header: initialized GPT header object
        """
        # zero the old checksum before calculating
        header.header_crc32 = 0
        # only CRC the header feilds 0x0 - 0x5b
        header.header_crc32 = binascii.crc32(header.marshal()[:92])



import json
import struct
import uuid
from enum import Enum
from math import ceil
from typing import List

from gpt_image.geometry import Geometry


class PartitionEntryError(Exception):
    """Exception class for erros in partition entries"""


class PartitionAttribute(Enum):
    """Supported partition attribute settings

    https://en.wikipedia.org/wiki/GUID_Partition_Table#Partition_entries_(LBA_2-33)


    The attribute field is 8 bytes with the last 2 bytes used for basic data. This
    field sets a bit in the last 2 bytes.
    """

    NONE = 0
    READ_ONLY = 60
    SHADOW_COPY = 61
    HIDDEN = 62
    NO_DRIVE_LETTER = 63


class Partition:
    """Partition class represents a GPT partition

    Start and end LBA are set to None because they must be calculated
    from a table's partition list.
    """

    _PARTITION_FORMAT = struct.Struct("<16s16sQQQ72s")
    _EMPTY_GUID = "00000000-0000-0000-0000-000000000000"
    # https://en.wikipedia.org/wiki/GUID_Partition_Table#Partition_entries
    LINUX_FILE_SYSTEM = "0FC63DAF-8483-4772-8E79-3D69D8477DE4"
    EFI_SYSTEM_PARTITION = "C12A7328-F81F-11D2-BA4B-00A0C93EC93B"

    def __init__(
        self,
        name: str,
        size: int,
        type_guid: str,
        partition_guid: str = "",
        alignment: int = 8,
        partition_attributes: int = 0,
    ):
        """Initialize Partition Object"""
        self.type_guid = type_guid
        self.partition_name = name
        self._attribute_flags = 0
        self.partition_guid = partition_guid
        # if the partition GUID is empty, generate one
        if not partition_guid:
            self.partition_guid = str(uuid.uuid4())
        self.first_lba = 0
        self.last_lba = 0
        self._attribute_flags = partition_attributes
        self.alignment = alignment
        self.size = size

    def __repr__(self) -> str:
        partitionvalue = {k: v for k, v in vars(self).items() if not k.startswith("_")}
        partitionvalue['attribute_flags'] = self.attribute_flags
        return json.dumps(partitionvalue, indent=2, ensure_ascii=False)

    @property
    def attribute_flags(self) -> List[int]:
        """Return a list of partition attribute flags

        Returns an empty list if no attributes are set

        """
        flags = []
        flag_int = self._attribute_flags
        while flag_int:
            flags.append((flag_int).bit_length() - 1)
            # remove the MSB by masking the integer by length -1
            flag_int = flag_int & ((1 << (flag_int.bit_length() - 1)) - 1)
        return flags

    @attribute_flags.setter
    def attribute_flags(self, flag: PartitionAttribute):
        """Set the partition attribute flag

        Sets the bit corresponding to the PartitionAttribute Class.
        If the PartitionAttribute.NONE value is set, this clears all flags
        """
        if flag.value == 0:
            self._attribute_flags = 0
        else:
            self._attribute_flags = self._attribute_flags | (1 << flag.value)

    def marshal(self) -> bytes:
        """Marshal to byte structure"""
        attributes_value = 0
        for attrib in self.attribute_flags:
            attributes_value += 2**attrib
        partition_bytes = self._PARTITION_FORMAT.pack(
            uuid.UUID(self.type_guid).bytes_le,
            uuid.UUID(self.partition_guid).bytes_le,
            self.first_lba,
            self.last_lba,
            attributes_value,
            bytes(self.partition_name, encoding="utf_16_le"),
        )
        return partition_bytes

    @staticmethod
    def read(partition_bytes: bytes, sector_size: int) -> "Partition":
        """Create a Partition object from existing bytes"""
        if len(partition_bytes) != PartitionEntryArray.EntryLength:
            raise ValueError(f"Invalid Partition Entry length: {len(partition_bytes)}")
        (
            type_guid,
            partition_guid,
            first_lba,
            last_lba,
            attribute_flags,
            partition_name,
        ) = Partition._PARTITION_FORMAT.unpack(partition_bytes)
        size = (last_lba - first_lba + 1) * sector_size

        part = Partition(
            partition_name.decode("utf_16_le").rstrip("\x00"),
            size,
            str(uuid.UUID(bytes_le=type_guid)),
            str(uuid.UUID(bytes_le=partition_guid)),
            partition_attributes=attribute_flags,
        )
        part.first_lba = first_lba
        part.last_lba = last_lba
        return part


class PartitionEntryArray:
    """Stores the Partition objects for a Table"""

    EntryCount = 128
    EntryLength = 128

    def __init__(self, geometry: Geometry):
        self.entries: List[Partition] = []
        self._geometry: Geometry = geometry

    def add(self, partition: Partition) -> None:
        """Add a partition to the entries

        Appends the Partition to the next available entry. Calculates the
        LBA's
        """
        partition.first_lba = self._get_first_lba(partition)
        partition.last_lba = self._get_last_lba(partition)
        self.entries.append(partition)

    def _get_first_lba(self, partition: Partition) -> int:
        """Calculate the first LBA of a new partition

        Search for the largest LBA, this will be used to calculate the first
        LBA of the partition being created.  If it is 0, all partitions are empty
        and the last lba is considered 33.

        The start sector (LBA) will take the alignment into account.

        """

        def next_lba(end_lba: int, alignment: int):
            m = int(end_lba / alignment)
            return (m + 1) * alignment

        largest_lba = 0
        for part in self.entries:
            lba = part.last_lba
            if int(lba) > int(largest_lba):
                largest_lba = lba
        last_lba = 33 if largest_lba == 0 else largest_lba
        return next_lba(int(last_lba), partition.alignment)

    def _get_last_lba(self, partition: Partition) -> int:
        """Calculate the last LBA of a new partition

        The last LBA will always be the -1 from the total partition LBA

        """
        if partition.size < self._geometry.sector_size:
            raise PartitionEntryError("Partition smaller than sector size")

        # round the LBA up to ensure our LBA will hold the partition
        lba = int(ceil(partition.size / self._geometry.sector_size))
        f_lba = int(partition.first_lba)
        return (f_lba + lba) - 1

    def marshal(self) -> bytes:
        """Convert the Partition Entry Array to its byte structure"""
        parts = [x.marshal() for x in self.entries]
        part_bytes = b"".join(parts)
        # pad the rest with zeros
        padded = part_bytes + b"\x00" * (
            (PartitionEntryArray.EntryCount * PartitionEntryArray.EntryLength)
            - len(part_bytes)
        )
        return padded



import json

import pytest

from gpt_image.disk import Disk
from gpt_image.partition import Partition

BYTE_DATA = b"\x01\x02\x03\x04"
DISK_SIZE = 2 * 1024 * 1024
OFFSET = 32


@pytest.fixture
def new_image(tmp_path):
    image_name = tmp_path / "test.img"
    disk = Disk(image_name)
    disk.create(DISK_SIZE * 2)
    part1 = Partition("partition1", 2 * 1024, Partition.LINUX_FILE_SYSTEM)
    part2 = Partition("partition2", 3 * 1024, Partition.LINUX_FILE_SYSTEM)
    disk.table.partitions.add(part1)
    disk.table.partitions.add(part2)
    disk.write()
    return image_name


def test_disk_create(tmp_path):
    """Test creating a new disk image"""
    disk_path = tmp_path / "test-disk.img"
    disk = Disk(disk_path)
    assert disk.image_path == disk_path.resolve()
    assert disk.sector_size == 512
    disk.create(DISK_SIZE)


def test_disk_open(new_image):
    disk = Disk(new_image)
    disk.open()
    assert disk.size == DISK_SIZE * 2
    assert disk.table.primary_header.backup is False
    assert disk.table.secondary_header.backup is True
    assert disk.table.primary_header.partition_entry_lba == 2
    assert disk.table.secondary_header.partition_entry_lba == 8159
    assert len(disk.table.partitions.entries) == 2
    assert disk.table.partitions.entries[0].partition_name == "partition1"


def test_disk_info(new_image):
    disk = Disk(new_image)
    disk.open()
    disk_s = str(disk)
    disk_d = json.loads(disk_s)
    assert not disk_d["primary_header"]["backup"]
    assert disk_d["backup_header"]["backup"]
    assert disk_d["backup_header"]["header_size"] == 92
    assert len(disk_d["partitions"]) == 2
